/*
 *		IP_MASQ_GG modu maskujcy do Gadu-Gadu
 *
 *
 * Wersja:	ip_masq_gg.c 1.0pre4   24/04/2003
 *
 * Autor:	Jakub Pidel <qbic@staszic.waw.pl>
 *
 * Bazowaem na:
 *	IP_MASQ_FTP Wouter'a Gadeyne
 * 	IP_MASQ_PORTFW Steven'a Clarke
 *
 * Obsuga wielu portw
 *	Modu moe nasuchiwa na wielu portach. Zwyczajowo s to 443
 *	(wiem, e to ssl ale powiedzcie to ludziom od GG) oraz 8074.
 *	Mona zmieni t list dopisujc insmod-owi parametr
 *		ports=x1,x2,x3...
 *	lub dopisujc to do /etc/modules.conf.
 *	Maksymalna liczba portw to MAX_MASQ_APP_PORTS, czyli zazwyczaj 12.
 *
 * Zabezpieczenie przed dziur "extended ALG vulnerability"
 *	Zabezpieczenie jest trywialne i polega na uniemoliwieniu 
 *	pocze na porty poniej 1024, oraz te ktre zdefiniuje uytkownik.
 *	Mona to uczyni parametrem noport=x1,x2,x3...
 *	Znw portw tych moe by MAX_MASQ_APP_PORTS, czyli zazwyczaj 12.
 *
 * Blokada pocze bezporednich
 *	Ustawienie parametru dcc=0 powoduje, i w pakiecie wysyanym
 *	do serwera GG adres i port pocze bezporednich jest ustawiany
 *	na 0 - znak, i klient ma poczenia bezp. wyczone.
 */

/*
 *  (C) Copyright 2003 by Jakub Pidel <qbic@staszic.waw.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <net/protocol.h>
#include <net/tcp.h>

/*
 * PARANOID - wczenie tej opcji powoduje, i wpuszczane s tylko poczenia
 * pochodzce z adresw nalecych do przyjaci (ludzi z listy).
 * Oficjalnie serwer GG rozsya adres tylko ludziom, ktrzy widniej na licie
 * znajomych, jednak kto moe zgadn ten adres lub bawi si portskanerem.
 * Niestety s te minusy tego rozwizania, jeli przyjaciel jest
 * "niewidzialny" nie dostaniemy jego IP przez co nie bdzie mia wstpu.
 * Problem take si pojawia kiedy kto ma wicej IPkw i z serwerem czy si
 * przez jeden interface, a z nami przez drugi.
 */
/* #define PARANOID */

/*
 * Jeli jest wczone PARANOID trzeba ograniczy ilo przyjaci jak
 * user moe posiada, gdy inaczej mgby wywrci kernel przez zajcie
 * caej pamici. :) Wystarczyoby tylko zarejestrowa sobie tak z 5 mln
 * przyjaci kady po 12 bajtw i 60MB pamici mamy z gowy...
 * Paru takich userw i system robi dup.
 */
#ifdef PARANOID
#define MAX_FRIENDS_COUNT	64
#endif

/*
 * INTERNAL_DCC_ROUTING - opcja ta wcza system podmieniania adresw
 * w powiadomieniach. Jeli przyjaciel z naszej listy korzysta z tej samej
 * maskarady co my, jego adres w powiadomieniu dla nas zostanie podmieniony
 * na oryginalny (podobnie jak nasz adres w powiadomieniu dla niego) przez
 * co poczenie midzy nami bdzie przebiega bez udziau maskarady.
 * Efekt jest taki, e my widzimy przyjaciela pod adresem wew. kiedy reszta
 * wiata widzi adres zew.
 * Moe powodowa problemy jeli router nie dzieli sieci na LAN i INET, ale
 * na wicej czci wzajemnie przed sob zamaskowanych.
 * Z drugiej jednak strony w typowym przypadku LAN/INET wyczenie tej opcji
 * uczyni poczenie niewykonalnym, gdy jedna strona bdzie wysya pakiety
 * na adres zew. drugiej, a druga na adres wew. pierwszej (ze swojego adresu
 * wew.).
 */
#define INTERNAL_DCC_ROUTING

/* #define CONFIG_IP_MASQ_DEBUG */

#include <net/ip_masq.h>
#include <net/ip_masq_mod.h>

/*
 * Zakres portw przydzielanych dla dcc.
 * Jak kto ma elegantszy sposb na alokacj portw, prosz o kontakt.
 * Pocztek przedziau to koniec przedziau naturalnego dla maskarady,
 * a koniec 1024 numerkw dalej.
 * Zakresu tego nie mona ustawi poniej 6000. portu bo bdzie kolizja
 * z automatem przydzielajcym porty dla u-space. Ciekawe efekty moe
 * powodowa take pokrycie si tej przestrzeni z przestrzeni maskarady ;).
 */
#define PORT_MASQ_GG_BEGIN	PORT_MASQ_END
#define PORT_MASQ_GG_END	PORT_MASQ_GG_BEGIN + 1024

/*
 * Konwersja zapisu liczb-sw z GG do naturalnego dla naszego kompa
 * i z powrotem. GG jest LITTLE-ENDIAN.
 */
#ifdef __BIG_ENDIAN
#define htogs(x)	swab16((x))
#define htogl(x)	swab32((x))
#define gtohs(x)	swab16((x))
#define gtohl(x)	swab32((x))
#else
#define htogs(x)	(x)
#define htogl(x)	(x)
#define gtohs(x)	(x)
#define gtohl(x)	(x)
#endif

/*
 * Wielko tablicy hashujcej poczenia bezp. Powinna by liczb pierwsz.
 */
#define IP_MASQ_GG_HASH	127

/*
 * Flagi pakietw w protokole GG. Wicej dokumentacji w ekg.
 */
#define GG_LOGIN	0x000c
#define GG_LOGIN_OK	0x0003
#define GG_NEW_STATUS	0x0002
#define GG_NOTIFY_REPLY	0x000c
#define GG_REM_NOTIFY	0x000e
#define GG_SEND_MSG	0x000b
#define GG_RECV_MSG	0x000a
#define GG_DISCONNECT	0x000b

#define GG_STATUS_NOT_AVAIL		0x0001
#define GG_STATUS_NOT_AVAIL_DESCR	0x0015
#define GG_STATUS_AVAIL_DESCR		0x0004
#define GG_STATUS_BUSY_DESCR		0x0005
#define GG_STATUS_INVISIBLE_DESCR	0x0016

/*
 * Linux ma domylne wyrwnanie do 32bitw, a niektre pakiety maj
 * rozmiar niepodzielny przez 4. Niestety np. ekg tego nie uwzgldnia.
 */
#define SIZEOF_GG_HEADER	8
#define SIZEOF_GG_LOGIN		22
#define SIZEOF_GG_NEW_STATUS	4
#define SIZEOF_GG_NOTIFY_REPLY	20
#define SIZEOF_GG_SEND_MSG	12
#define SIZEOF_GG_RECV_MSG	16
#define SIZEOF_GG_REM_NOTIFY	5

/*
 * struct gg_user - Struktura przypisywana poszczeglnym userom.
 * Powstaje przy pierwszym logowaniu, a wyrzucana jest przy
 * wyadowywaniu moduu.
 * Kiedy user jest zalogowany, *conn wskazuje na struktur poczenia.
 */
struct gg_user {
	__u32 uin;
	__u32 status;
	__u32 version;
	__u32 recv_msgs;
	__u32 sent_msgs;
	struct gg_connection *conn;
	struct gg_user *next;
};

/*
 * struct gg_connection - Struktura poczenia. Powstaje dla kadego
 * poczenia gwnego (klient gg - serwer gg).
 * saddr i maddr - adresy odpowiednio rzeczywisty i widoczny z zew. dla
 *   danego poczenia. Dana na stae BIG-ENDIAN (typowo dla protokou IP).
 * active - to flaga okrelajca czy wpuszcza poczenia bezp.
 *   Kiedy klient wysa pakiet logujcy, ale nie dosta jeszcze odpowiedzi
 *   LOGIN_OK, lub kiedy si wylogowa, ale timeout poczenia jeszcze
 *   nie upyn, flaga ta jest ustawiona na 0.
 */
struct gg_connection {
#ifdef PARANOID
	struct gg_friend *friends;
	int friends_count;
#endif
	struct gg_user *user;
	struct gg_connection *next;
	unsigned int id;
	__u32 saddr, maddr;
	__u16 sport, mport;
	__u8 active;
};

/*
 * struct gg_friend - struktura budujca list przyjaci.
 */
#ifdef PARANOID
struct gg_friend {
	struct gg_friend *next;
	__u32 uin;
	__u32 addr;
};
#endif

/*
 * Bardziej zoone pakiety protokou GG. Wicej informacji w dokumentacji
 * do ekg.
 */
struct gg_header {
	__u32 type;
	__u32 length;
};

struct gg_login {
	__u32 uin;
	__u32 hash;
	__u32 status;
	__u32 version;
	__u32 local_ip;
	__u16 local_port;
};

struct gg_notify_reply {
	__u32 uin;
	__u32 status;
	__u32 remote_ip;
	__u16 remote_port;
	__u16 version_lo;
	__u16 version_hi;
	__u16 unknown1;
};

/*
 * Lista userw, ktrzy si do tej pory czyli.
 */
static struct gg_user *users = NULL;

/*
 * Tablica hashujca poczenia.
 */
static struct gg_connection *connections[IP_MASQ_GG_HASH];

/*
 * Zamki dla zachowania synchronizacji w operacjach na listach.
 */
static rwlock_t gg_user_lock = RW_LOCK_UNLOCKED;
static rwlock_t gg_conn_lock = RW_LOCK_UNLOCKED;
static rwlock_t gg_id_lock = RW_LOCK_UNLOCKED;
static rwlock_t gg_masq_lock = RW_LOCK_UNLOCKED;

/*
 * id - numer identyfikacyjny poczenia.
 * last_reg_port - ostatnio przydzielony port dla DCC, rotuje si cyklicznie
 *   po wartociach PORT_MASQ_GG_BEGIN - PORT_MASQ_GG_END
 */
static unsigned int id = 0;
static __u16 last_reg_port = PORT_MASQ_GG_BEGIN;

static struct ip_masq_mod *mmod_self = NULL;

/*
 * ports - lista portw na ktrych modu prowadzi nasuch.
 * masg_gg_objs - korespondujce z nimi struktury ip_masg_app.
 */
static int ports[MAX_MASQ_APP_PORTS] = { 443, 8074 };
static struct ip_masq_app *masq_gg_objs[MAX_MASQ_APP_PORTS];

/*
 * noport - lista portw na ktre nie bd przekierowywane poczenia bezp.
 */
static int noport[MAX_MASQ_APP_PORTS];

/*
 * Poziom debug-a
 */
#ifdef CONFIG_IP_MASQ_DEBUG
static int debug = 0;
MODULE_PARM(debug, "i");
#endif

static int dcc = 1;
MODULE_PARM(dcc, "i");

MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_MASQ_APP_PORTS) "i");
MODULE_PARM(noport, "1-" __MODULE_STRING(MAX_MASQ_APP_PORTS) "i");

/*
 * gg_reg_port() - alokuj i rejestruj port zew. dla nowego poczenia.
 * Parametry:
 *   conn - struktura poczenia, ktr zamierzamy umieci w connections[].
 *   maddr - zew. adres jakiego uywa maskarada.
 * Zwracane wartoci:
 *   jeli port zostanie przydzielony wewntrz struktury conn ustawiane
 *   s pola maddr i mport, oraz modyfikowane jest pole next
 *   (aby lista portw miaa sens).
 */
static __inline__ void gg_reg_port(struct gg_connection *conn, __u32 maddr) {
	unsigned int hash, port;
	struct gg_connection *ptr;
	
	write_lock_bh(&gg_conn_lock);

	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_reg_port()...\n", conn->id);

	port = last_reg_port;
	do {
		port++;
		if (port == PORT_MASQ_GG_END || port >= 65536)
			port = PORT_MASQ_GG_BEGIN;
		
		hash = port % IP_MASQ_GG_HASH;
		ptr = connections[hash];
		while (ptr != NULL && ptr->mport != port)
			ptr = ptr->next;
		if (ptr == NULL) {
			IP_MASQ_DEBUG(1-debug, "GG[%d]: Registering DCC port %d.\n",
				conn->id, port);
			last_reg_port = port;
			conn->mport = port;
			conn->maddr = maddr;
			conn->next = connections[hash];
			connections[hash] = conn;
		/* Zaznacz, e kto ju z tego ustrojstwa korzysta */
			ip_masq_mod_inc_nent(mmod_self);
		}
	} while (port != last_reg_port);

	write_unlock_bh(&gg_conn_lock);
}

/*
 * gg_unreg_port() - dealokacja zew. portu.
 * Parametry:
 *   conn - struktura ktr chcemy wypi z listy w connections[]
 * Zwracane wartoci:
 *   pola maddr, mport i next s zerowane.
 */
static __inline__ void gg_unreg_port(struct gg_connection *conn) {
	struct gg_connection **ptr;
	
	write_lock_bh(&gg_conn_lock);

	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_unreg_port()...\n", conn->id);
	
	ptr = &connections[conn->mport % IP_MASQ_GG_HASH];
	while (*ptr != NULL) {
		if ((*ptr)->mport == conn->mport) {
			IP_MASQ_DEBUG(1-debug, "GG[%d]: Unregistering DCC port %d.\n",
				conn->id, conn->mport);
			*ptr = (*ptr)->next;
			conn->maddr = 0;
			conn->mport = 0;
			conn->next = NULL;
		/* Ustaw, e wanie stracilimy jednego klienta */
			ip_masq_mod_dec_nent(mmod_self);
			break;
		}
		ptr = &(*ptr)->next;
	}

	write_unlock_bh(&gg_conn_lock);
}

/*
 * gg_port_lookup() - sprawd czy para adres, port jest zarejestrowana
 *   w connections[].
 * Parametry:
 *   addr, port - interesujcy nas adres i port.
 * Zwracane wartoci:
 *   jak znajdzie - wskanik na struktur poczenia,
 *   jak nie to NULL.
 */
static struct gg_connection * gg_port_lookup(__u32 addr, __u16 port) {
	struct gg_connection *ptr;
	
	read_lock_bh(&gg_conn_lock);

	ptr = connections[port % IP_MASQ_GG_HASH];
	while (ptr != NULL && (ptr->maddr != addr || ptr->mport != port))
		ptr = ptr->next;
	
	read_unlock_bh(&gg_conn_lock);
	
	IP_MASQ_DEBUG(4-debug, "GG: gg_port_lookup(%d.%d.%d.%d, %d) %s.\n",
			NIPQUAD(addr), port,
			ptr == NULL ? "negative" :
			ptr->active ? "positive" :
			"inactive");
	return ptr;
}

/*
 * gg_reg_user() - znajd w rejestrze lub zarejestruj nowego usera.
 * Parametry:
 *   uin - numerek naszego uytkownika.
 * Zwracane wartoci:
 *   wskanik na struktur tego usera.
 */
static __inline__ struct gg_user * gg_reg_user(__u32 uin) {
	struct gg_user **ptr, *temp;
	
	write_lock_bh(&gg_user_lock);
	
	IP_MASQ_DEBUG(3-debug, "GG: gg_reg_user(%d)...\n", uin);
	
	ptr = &users;
	while (*ptr != NULL && (*ptr)->uin < uin)
		ptr = &(*ptr)->next;
	if (*ptr == NULL || (*ptr)->uin > uin) {
		IP_MASQ_DEBUG(1-debug, "GG: Registering user %d.\n", uin);
		if ((temp = kmalloc(sizeof(struct gg_user), GFP_ATOMIC)) == NULL)
			IP_MASQ_ERR("GG: No memory for user data!\n");
		memset(temp, 0, sizeof(struct gg_user));
		temp->uin = uin;
		temp->next = *ptr;
		*ptr = temp;
	}
	
	write_unlock_bh(&gg_user_lock);
	
	return *ptr;
}

/*
 * gg_valid_port() - sprawd czy port dcc klienta jest poprawny, czy
 * przypadkiem kto nie robi czego wrednego wykorzystujc nas do ustawienia
 * forwardu na port lokalnej usugi (np. ssh, samba itd.)
 * Parametry:
 *   port - numer portu.
 * Zwracane wartoci:
 *   prawda(1) lub fasz(0) - czy port jest ok.
 */
static int gg_valid_port(__u16 port) {
	int i;
	
	if (port < 1024)
		return 0;
	for (i = 0; i < MAX_MASQ_APP_PORTS; i++)
		if (port == noport[i])
			return 0;
	return 1;
}

#ifdef PARANOID
/*
 * gg_reg_friend() - zarejestruj nowego przyjaciela.
 * Parametry:
 *   conn - struktura naszego poczenia,
 *   uin - numerek przyjaciela,
 *   addr - jego adres (jesli == 0 to znaczy ze przyjaciel sobie poszed).
 * Zwracane wartoci:
 *   brak
 */
static void gg_reg_friend(struct gg_connection *conn, __u32 uin, __u32 addr) {
	struct gg_friend **friends, *ptr;
	
	write_lock_bh(&gg_conn_lock);

	friends = &conn->friends;
	while (*friends != NULL && (*friends)->uin != uin)
		friends = &(*friends)->next;
		
	if (addr != 0 && *friends != NULL) {
		IP_MASQ_DEBUG(1-debug, "GG[%d]: Friend entry edit: "
			"uin=%d addr=%d.%d.%d.%d\n", conn->id, uin, NIPQUAD(addr));
		(*friends)->addr = addr;
	} else
	if (addr != 0 && *friends == NULL && conn->friends_count < MAX_FRIENDS_COUNT) {
		IP_MASQ_DEBUG(1-debug, "GG[%d]: Registering friend: "
			"uin=%d addr=%d.%d.%d.%d\n", conn->id, uin, NIPQUAD(addr));
		if ((*friends = kmalloc(sizeof(struct gg_friend), GFP_ATOMIC)) == NULL)
			IP_MASQ_ERR("GG[%d]: No memory for friend data!\n", id);
		(*friends)->uin = uin;
		(*friends)->addr = addr;
		(*friends)->next = NULL;
		conn->friends_count++;
	} else
	if (addr == 0 && *friends != NULL) {
		ptr = *friends;
		IP_MASQ_DEBUG(1-debug, "GG[%d]: Unregistering friend: "
			"uin=%d addr=%d.%d.%d.%d\n", conn->id, uin, NIPQUAD(ptr->addr));
		*friends = (*friends)->next;
		conn->friends_count--;
		kfree_s(ptr, sizeof(struct gg_friend));
	}
	
	write_unlock_bh(&gg_conn_lock);
}

/*
 * gg_friend_lookup() - sprawdz czy przyjaciel
 * Parametry:
 *   conn - struktura naszego poczenia,
 *   addr - interesujcy nas adres.
 * Zwracane wartoci:
 *   prawda lub fasz - przyjaciel czy nie.
 */
static __inline__ int gg_friend_lookup(struct gg_connection *conn, __u32 addr) {
	struct gg_friend *friends;
	
	read_lock_bh(&gg_conn_lock);
	
	friends = conn->friends;
	while (friends != NULL && friends->addr != addr)
		friends = friends->next;
	
	read_unlock_bh(&gg_conn_lock);
	
	IP_MASQ_DEBUG(4-debug, "GG[%d]: gg_friend_lookup(%d.%d.%d.%d) %s.\n",
		conn->id, NIPQUAD(addr), friends != NULL ? "positive" : "negative");
	
	return friends != NULL;
}
#endif

/*
 * gg_init() - utwrz struktur poczenia i przydziel id.
 *   Uruchamiana za kadym razem kiedy pojawia si poczenie wychodzce
 *   na interesujce nas porty. Rwnie wtedy kiedy kto czy si z https-em.;)
 * Parametry:
 *   typowo dla takich funkcji
 * Zwracane wartoci:
 *   zawsze 0
 */
static int gg_init(struct ip_masq_app *mapp, struct ip_masq *ms) {

	write_lock_bh(&gg_id_lock);
	
	id++;
	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_init()...\n", id);
	
	if ((ms->app_data = kmalloc(sizeof(struct gg_connection), GFP_ATOMIC)) == NULL)
		IP_MASQ_ERR("GG[%d]: No memory for application data!\n", id);
	
	memset(ms->app_data, 0, sizeof(struct gg_connection));
	((struct gg_connection *)ms->app_data)->id = id;
	
	write_unlock_bh(&gg_id_lock);
	
	MOD_INC_USE_COUNT;
	return 0;
}

/*
 * gg_done() - wyrejestruj (jeli trzeba) oraz usu struktur poczenia.
 *   Uruchamiane jest kiedy upynie timeout poczenia w maskaradzie.
 * Parametry:
 *   typowo dla takich funkcji
 * Zwracane wartoci:
 *   zawsze 0
 */
static int gg_done(struct ip_masq_app *mapp, struct ip_masq *ms) {
#ifdef PARANOID
	struct gg_friend *ptr, *friends;
#endif

	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_done()...\n",
		((struct gg_connection *)ms->app_data)->id);

/* Jeli port by zarejestrowany to go wyrejestruj */
	if (((struct gg_connection *)ms->app_data)->mport != 0)
		gg_unreg_port((struct gg_connection *)ms->app_data);

/* Odcz struktur od struktury uytkownika */
	if (((struct gg_connection *)ms->app_data)->user != NULL
	&& ((struct gg_connection *)ms->app_data)->user->conn == ms->app_data)
		((struct gg_connection *)ms->app_data)->user->conn = NULL;

#ifdef PARANOID
/* Wywal list przyjaci */
	friends = ((struct gg_connection *)ms->app_data)->friends;
	while (friends != NULL) {
		ptr = friends;
		friends = friends->next;
		kfree_s(ptr, sizeof(struct gg_friend));
	}
#endif
	kfree_s(ms->app_data, sizeof(struct gg_connection));
	
	MOD_DEC_USE_COUNT;
	return 0;
}

/*
 * gg_out() - funkcja przechwytujca wszystkie pakiety wychodzce w obrbie
 *   interesujcego nas poczenia (ms).
 * Prarametry:
 *   skb_p - struktura pakietu,
 *   maddr - adres ifaceu, przez ktry zostanie wysany pakiet,
 *   reszta typowo dla takich funkcji.
 * Zwracane wartoci:
 *   zawsze 0
 */
int gg_out(struct ip_masq_app *mapp, struct ip_masq *ms,
		 struct sk_buff **skb_p, __u32 maddr) {
	struct sk_buff *skb;
	struct iphdr *iph;
	struct tcphdr *th;
	struct gg_header *header;
	struct gg_connection *conn;
	char *packet, *packet_end, *data;
	int data_len;

	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_out()...\n",
		((struct gg_connection *)ms->app_data)->id);

	/* Interesuj nas tylko nawizane poczenia,... */
	if (ms->state != IP_MASQ_S_ESTABLISHED)
		return 0;

	skb = *skb_p;
	iph = skb->nh.iph;
	th = (struct tcphdr *)&(((char *)iph)[iph->ihl * 4]);
	packet = (char *)th + (((struct tcphdr *)th)->doff << 2);
	packet_end = skb->h.raw + skb->len;
	conn = (struct gg_connection *)ms->app_data;
	
	while (packet + SIZEOF_GG_HEADER <= packet_end) {
		header = (struct gg_header *)packet;
		data = packet + SIZEOF_GG_HEADER;
		data_len = gtohl(header->length);
		
	/* ... zawierajace komplet danych */
		if (data + data_len > packet_end)
			return 0;
		
	/* User musi by zalogowany lub wanie si logowa */
		if (conn->user == NULL && gtohl(header->type) != GG_LOGIN)
			return 0;
	
		IP_MASQ_DEBUG(2-debug, "GG[%d]==> Packet: "
			"type=0x%08X length=%d\n",
			conn->id, gtohl(header->type),
			gtohl(header->length));
	
		switch (gtohl(header->type)) {
		case GG_LOGIN:
			if (data_len < SIZEOF_GG_LOGIN)
				return 0;
			
			if (conn->user != NULL)
				conn->user->conn = NULL;
			
			conn->user = gg_reg_user(((struct gg_login *)data)->uin);
			
			conn->user->status =
				gtohl(((struct gg_login *)data)->status);
			conn->user->version =
				gtohl(((struct gg_login *)data)->version);
			conn->saddr =
				((struct gg_login *)data)->local_ip;
			conn->sport =
				gtohs(((struct gg_login *)data)->local_port);
			conn->active = 0;
			
			if (ms->saddr != conn->saddr) {
				conn->saddr = ms->saddr;
				conn->sport = 0;
			}
			
			if (conn->mport != 0 && !gg_valid_port(conn->sport))
				gg_unreg_port(conn);
			
			if (dcc && conn->mport == 0 && gg_valid_port(conn->sport))
				gg_reg_port(conn, maddr);
			
			((struct gg_login *)data)->local_ip = conn->maddr;
			((struct gg_login *)data)->local_port = htogs(conn->mport);
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]==> Logging in: "
				"uin=%d status=0x%08X version=0x%08X "
				"saddr=%d.%d.%d.%d:%d "
				"maddr=%d.%d.%d.%d:%d\n",
				conn->id,
				conn->user->uin,
				conn->user->status,
				conn->user->version,
				NIPQUAD(conn->saddr),
				conn->sport,
				NIPQUAD(conn->maddr),
				conn->mport);
			break;

		case GG_NEW_STATUS:
			if (data_len < SIZEOF_GG_NEW_STATUS)
				return 0;
			
			conn->user->status = gtohl(*((__u32 *)data));
			
			if ((conn->user->status & 0x000000ff) 
					== GG_STATUS_NOT_AVAIL
				|| (conn->user->status & 0x000000ff)
					== GG_STATUS_NOT_AVAIL_DESCR)
				conn->active = 0;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]==> New status=0x%08X\n",
				conn->id,
				conn->user->status);
			break;
		
		case GG_SEND_MSG:
			if (data_len < SIZEOF_GG_SEND_MSG)
				return 0;
				
			conn->user->sent_msgs++;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]==> Message to "
				"uin=%d seq=%d class=0x%08X\n",
				conn->id,
				gtohl(((__u32 *)data)[0]),
				gtohl(((__u32 *)data)[1]),
				gtohl(((__u32 *)data)[2]));
			break;
#ifdef PARANOID
		case GG_REM_NOTIFY:
			if (data_len < SIZEOF_GG_REM_NOTIFY)
				return 0;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]==> Remove notify: "
				"uin=%d type=0x%02X\n",
				conn->id, gtohl(((__u32 *)data)[0]), ((char *)data)[4]);
				
			if (((char *)data)[4] == 0x03)
				gg_reg_friend(conn, gtohl(((__u32 *)data)[0]), 0);
			break;
#endif
		}
		packet += SIZEOF_GG_HEADER + data_len;
	}
	return 0;
}

/*
 * gg_in() - funkcja przechwytujca wszystkie pakiety przychodzce w obrbie
 *   interesujcego nas poczenia (ms).
 * Prarametry:
 *   skb_p - struktura pakietu,
 *   maddr - adres ifaceu, przez ktry zosta odebrany pakiet,
 *   reszta typowo dla takich funkcji.
 * Zwracane wartoci:
 *   zawsze 0
 */
int gg_in(struct ip_masq_app *mapp, struct ip_masq *ms,
		 struct sk_buff **skb_p, __u32 maddr) {
	struct sk_buff *skb;
	struct iphdr *iph;
	struct tcphdr *th;
	struct gg_header *header;
	struct gg_connection *conn, *ptr;
	struct gg_notify_reply *reply;
	__u32 status;
	char *packet, *packet_end, *data;
	int data_len, reply_len;

	IP_MASQ_DEBUG(3-debug, "GG[%d]: gg_in()...\n",
		((struct gg_connection *)ms->app_data)->id);

	/* Interesuj nas tylko nawizane poczenia,... */
	if (ms->state != IP_MASQ_S_ESTABLISHED)
		return 0;

	skb = *skb_p;
	iph = skb->nh.iph;
	th = (struct tcphdr *)&(((char *)iph)[iph->ihl * 4]);
	packet = (char *)th + (((struct tcphdr *)th)->doff << 2);
	packet_end = skb->h.raw + skb->len;
	conn = (struct gg_connection *)ms->app_data;
	
	while (packet + SIZEOF_GG_HEADER <= packet_end) {
		header = (struct gg_header *)packet;
		data = packet + SIZEOF_GG_HEADER;
		data_len = gtohl(header->length);
		
	/* ... zawierajace komplet danych */
		if (data + data_len > packet_end)
			return 0;
		
	/* User musi by zalogowany*/
		if (conn->user == NULL)
			return 0;

		IP_MASQ_DEBUG(2-debug, "GG[%d]<== Packet: "
			"type=0x%08X length=%d\n",
			conn->id,
			gtohl(header->type),
			gtohl(header->length));

		switch (gtohl(header->type)) {
		case GG_LOGIN_OK:
			if (data_len != 0)
				return 0;
			
			conn->active = 1;
			conn->user->conn = conn;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]<== Login ok.\n",
				conn->id);
			
			break;

		case GG_NOTIFY_REPLY:
			reply = (struct gg_notify_reply *)data;
			reply_len = data_len;
			while (reply_len >= SIZEOF_GG_NOTIFY_REPLY) {
				status = gtohl(reply->status) & 0x000000ff;
				ptr = NULL;
				
				IP_MASQ_DEBUG(1-debug, "GG[%d]<== Notify reply: "
					"uin=%d status=0x%08X version=0x%04X%04X "
					"addr=%d.%d.%d.%d:%d\n",
					conn->id,
					gtohl(reply->uin),
					gtohl(reply->status),
					gtohs(reply->version_hi),
					gtohs(reply->version_lo),
					NIPQUAD(reply->remote_ip),
					gtohs(reply->remote_port));
#ifdef INTERNAL_DCC_ROUTING
				ptr = gg_port_lookup(reply->remote_ip,
					gtohs(reply->remote_port));
				if (ptr != NULL) {
					reply->remote_ip = ptr->saddr;
					reply->remote_port = htogs(ptr->sport);
					IP_MASQ_DEBUG(1-debug, "GG[%d]^^^ "
						"Changed to addr=%d.%d.%d.%d:%d\n",
						conn->id,
						NIPQUAD(ptr->saddr),
						ptr->sport);
				}
#endif
#ifdef PARANOID
				if (ptr == NULL)
					gg_reg_friend(conn, gtohl(reply->uin),
						reply->remote_ip);
#endif
			/* status z opisem musi by ostatni w pakiecie */
				if (status == GG_STATUS_NOT_AVAIL_DESCR
					|| status == GG_STATUS_AVAIL_DESCR
					|| status == GG_STATUS_BUSY_DESCR
					|| status == GG_STATUS_INVISIBLE_DESCR)
					break;
				
				reply++;
				reply_len -= SIZEOF_GG_NOTIFY_REPLY;
			}
			break;

		case GG_RECV_MSG:
			if (data_len < SIZEOF_GG_RECV_MSG)
				return 0;
			
			conn->user->recv_msgs++;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]<== Message from "
				"uin=%d seq=%d time=%d class=0x%08X\n",
				conn->id,
				gtohl(((__u32 *)data)[0]),
				gtohl(((__u32 *)data)[1]),
				gtohl(((__u32 *)data)[2]),
				gtohl(((__u32 *)data)[3]));
			break;
			
		case GG_DISCONNECT:
			if (data_len != 0)
				return 0;
			
			conn->active = 0;
			
			IP_MASQ_DEBUG(1-debug, "GG[%d]<== Disconnect\n",
				conn->id);
			break;
		}
		packet += SIZEOF_GG_HEADER + data_len;
	}
	return 0;
}

#ifdef CONFIG_PROC_FS
/*
 * gg_procinfo() - funkcja generujca zawarto pliku w profsie.
 * Parametry:
 *   buffer - bufor na dane,
 *   start - pocztek danych w buforze,
 *   offset - "chcemy dane od offsetu...",
 *   length - ile tych danych chcemy.
 * Zwracane wartoci:
 *   ile danych udao si wygenerowa.
 */
static int gg_procinfo(char *buffer, char **start, off_t offset, int length,
		int unused) {
	char temp[128], saddr[22], maddr[22];
	char *pos = buffer;
	struct gg_user *user;
	int x, len = 0;
	
	read_lock_bh(&gg_user_lock);
	read_lock_bh(&gg_conn_lock);
	
	IP_MASQ_DEBUG(3-debug, "GG: gg_procinfo()...\n");
	
	if (offset < 94) {
		strncpy(temp, "  GG UIN      Status    Version    SentMsgs   "
			"RecvMsgs        Real Addr             Masq Addr\n", 128);
		x = length - len < 94 - offset ? length - len : 94 - offset;
		memcpy(pos, temp + offset, x);
		offset = 0;
		len += x;
		pos += x;
	} else
		offset -= 94;
	
	for (user = users; user != NULL && length > len; user = user->next) {
		if (offset < 100) {
		/* jeli user niezalogowany */
			strncpy(saddr, "---.---.---.---:-----", 22);
			strncpy(maddr, "---.---.---.---:-----", 22);
			if (user->conn != NULL && user->conn->active) {
		/* jeli zalogowany */
				sprintf(saddr, "%d.%d.%d.%d:%-5d",
					NIPQUAD(user->conn->saddr),
					user->conn->sport);
				sprintf(maddr, "%d.%d.%d.%d:%-5d",
					NIPQUAD(user->conn->maddr),
					user->conn->mport);
			}
			sprintf(temp, "% 10d: 0x%08X 0x%08X % 10d % 10d %21s %21s\n",
				user->uin, user->status, user->version,
				user->sent_msgs, user->recv_msgs, saddr, maddr);
			x = length - len < 100 - offset
				? length - len : 100 - offset;
			memcpy(pos, temp + offset, x);
			offset = 0;
			len += x;
			pos += x;
		} else
			offset -= 100;
	}
	
	read_unlock_bh(&gg_conn_lock);
	read_unlock_bh(&gg_user_lock);
	
	*start = buffer;
	return len;
}
#endif

/*
 * gg_dcc_rule() - czy wolno pakiet dcc przepuci do rodka.
 * Parametry:
 *   typowo dla takich funkcji.
 * Zwracane wartoci:
 *   prawda lub fasz.
 */
static int gg_dcc_rule(const struct sk_buff *skb, const struct iphdr *iph) {
	struct tcphdr *th = (struct tcphdr *)&(((char *)iph)[iph->ihl * 4]);

	IP_MASQ_DEBUG(4-debug, "GG: gg_rule()...\n");
	
	return (iph->protocol == IPPROTO_TCP
		&& gg_port_lookup(iph->daddr, ntohs(th->dest)) != NULL);
}

/*
 * gg_dcc_create() - utwrz poczenie przychodzce dcc.
 *   Jeli wczone PARANOID poza sprawdzeniem, czy port jest rzeczywicie
 *   przekierowany, sprawd czy to przyjaciel.
 * Parametry:
 *   typowo dla takich funkcji.
 * Zwracane wartoci:
 *   typowo dla takich funkcji.
 */
static struct ip_masq * gg_dcc_create(const struct sk_buff *skb,
		const struct iphdr *iph, __u32 maddr) {
	struct tcphdr *th = (struct tcphdr *)&(((char *)iph)[iph->ihl * 4]);
	struct gg_connection *ptr;
	struct ip_masq *ms = NULL;
	
	write_lock(&gg_masq_lock);
	IP_MASQ_DEBUG(4-debug, "GG: gg_create()...\n");
	
	if (iph->protocol == IPPROTO_TCP
		&& (ptr = gg_port_lookup(iph->daddr, ntohs(th->dest))) != NULL
		&& ptr->active
#ifdef PARANOID
		&& gg_friend_lookup(ptr, iph->saddr)
#endif
		) {
		IP_MASQ_DEBUG(1-debug, "GG[%d]: DCC redirect "
			"saddr=%d.%d.%d.%d:%d maddr=%d.%d.%d.%d:%d "
			"daddr=%d.%d.%d.%d:%d\n",
			ptr->id,
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(ptr->maddr), ptr->mport,
			NIPQUAD(ptr->saddr), ptr->sport);
			
		ms = ip_masq_new(IPPROTO_TCP, iph->daddr, th->dest,
			ptr->saddr, htons(ptr->sport),
			iph->saddr, th->source, 0);
		
		if (ms != NULL)
			ip_masq_listen(ms);
	}
	
	write_unlock(&gg_masq_lock);
	
	return ms;
}

struct ip_masq_app ip_masq_gg_app = {
	NULL,			/* next */
	"gg",			/* name */
	0,			/* type */
	0,			/* n_attach */
	gg_init,
	gg_done,
	gg_out,
	gg_in,
};

#ifdef CONFIG_PROC_FS
static struct proc_dir_entry ip_masq_gg_proc_entry = {
	0,			/* low_ino */
	2,			/* namelen */
	"gg",			/* name */
	S_IFREG | S_IRUGO,	/* mode */
	1,			/* nlink */
	0,			/* uid */
	0,			/* gid */
	0,			/* size */
	&proc_net_inode_operations,
	gg_procinfo
};
#define ip_masq_gg_proc_ent	&ip_masq_gg_proc_entry
#else
#define ip_masq_gg_proc_ent	NULL
#endif

static struct ip_masq_mod ip_masq_gg_mod = {
	NULL,			/* next */
	NULL,			/* next_reg */
	"gg",			/* name */
	ATOMIC_INIT(0),		/* nent */
	ATOMIC_INIT(0),		/* refcnt */
	ip_masq_gg_proc_ent,	/* procfs entry */
	NULL,			/* u-space function */
	NULL,			/* masq_mod_init */
	NULL,			/* masq_mod_done */
	gg_dcc_rule,
	NULL,
	gg_dcc_create,
	NULL,
	NULL,
	NULL,
};

/*
 * ip_masq_gg() - funkcja inicjujca modu. Rejestruj nasuch, przekierowanie
 *   pakietw i takie tam.
 */
__initfunc(int ip_masq_gg_init(void)) {
	struct ip_masq_app *mapp;
	int i, ret = 0, err;
	
	IP_MASQ_DEBUG(3-debug, "GG: ip_masq_gg_init()...\n");
	
	if ((err = register_ip_masq_mod(mmod_self = &ip_masq_gg_mod)) != 0)
		return err;
	
	for (i = 0; i < MAX_MASQ_APP_PORTS; i++) {
		masq_gg_objs[i] = NULL;
		if (ports[i]) {
			mapp = kmalloc(sizeof(struct ip_masq_app), GFP_KERNEL);
			if (mapp == NULL)
				return -ENOMEM;
			memcpy(mapp, &ip_masq_gg_app, sizeof(struct ip_masq_app));
			
			if ((err = ip_masq_app_init_proto_port(mapp,
				IP_MASQ_APP_OUTBOUND, IPPROTO_TCP, ports[i])) != 0
			|| (err = register_ip_masq_app_type(mapp)) != 0) {
				ret = err;
				kfree_s(mapp, sizeof(struct ip_masq_app));
				break;
			}
			
			masq_gg_objs[i] = mapp;
			IP_MASQ_DEBUG(1-debug, "GG: Loaded support on port %d.\n",
				ports[i]);
		}
	}
	return ret;
}

/*
 * ip_masq_gg() - funkcja koczca prac moduu. Wyrejestruj wszystko.
 */
int ip_masq_gg_done(void) {
	struct gg_user *ptr;
	int i, ret = 0, err;
	
	IP_MASQ_DEBUG(3-debug, "GG: ip_masq_gg_done()...\n");
	
	if ((err = unregister_ip_masq_mod(&ip_masq_gg_mod)) != 0)
		return err;
	
	for (i = 0; i < MAX_MASQ_APP_PORTS; i++) {
		if (masq_gg_objs[i]) {
			if ((err = unregister_ip_masq_app(masq_gg_objs[i])) != 0)
				ret = err;
			
			kfree_s(masq_gg_objs[i], sizeof(struct ip_masq_app));
			masq_gg_objs[i] = NULL;
			
			IP_MASQ_DEBUG(1-debug, "GG: Unloaded support on port %d.\n",
				ports[i]);
		}
	}
	
	while (users != NULL) {
		ptr = users;
		users = users->next;
		kfree_s(ptr, sizeof(struct gg_user));
	}
	
	return ret;
}

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void) {
	if (ip_masq_gg_init() != 0)
		return -EIO;
	return 0;
}

void cleanup_module(void) {
	if (ip_masq_gg_done() != 0)
		IP_MASQ_ERR("GG: can't remove module!\n");
}

#endif /* MODULE */
